function simulations = pruning_abounds(M_,options_,order,type)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% pruning_abounds.m
%
% This file implements all of the second and third order pruning algorithms
% documented in "Pruning in Perturbation DSGE Models" by Hong Lan and Alexander
% Meyer-Gohde.
%
% The pruning algorithms are called by setting the option type:
%
% For second order approximations (options_.order=2 in Dynare)
% type = 
%'kim_et_al' the second order algorithm of 
%               KIM, J., S. KIM, E. SCHAUMBURG, AND C. A. SIMS (2008): Calculating and Using Second-
%               Order Accurate Solutions of Discrete Time Dynamic Equilibrium Models, Journal of Economic
%               Dynamics and Control, 32(11), 33973414.
%
%'den_haan_de_wind' the second order algorithm of
%               DEN HAAN, W. J., AND J. DE WIND (2012): Nonlinear and Stable Perturbation-Based Approximations,
%               Journal of Economic Dynamics and Control, 36(10), 14771497.
%
%'lan_meyer-gohde' the second order algorithm of 
%               LAN, H., AND A. MEYER-GOHDE(Forthcoming): Solving DSGE Models with a Nonlinear Moving Average, Journal of
%               Economic Dynamics and Control.
%
%
% For third order approximations (options_.order=3 in Dynare)
% type = 
%'andreasen' the third order algorithm of 
%               ANDREASEN, M. M. (2012): On the Effects of Rare Disasters and Uncertainty Shocks for Risk
%               Premia in Non-Linear DSGE Models, Review of Economic Dynamics, 15(3), 295316.
%
%'fernandez-villaverde_et_al' the third order algorithm of
%               FERNANDEZ-VILLAVERDE, J., P. A. GUERRO N-QUINTANA, J. RUBIO-RAMI REZ, AND
%               M. URIBE (2011): Risk Matters: The Real Effects of Volatility Shocks, American Economic
%               Review, 101(6), 253061.
%
%'den_haan_de_wind' the third order algorithm of
%               DEN HAAN, W. J., AND J. DE WIND (2012): Nonlinear and Stable Perturbation-Based Approximations,
%               Journal of Economic Dynamics and Control, 36(10), 14771497.
%
%'lan_meyer-gohde' the third order algorithm of 
%               LAN, H., AND A. MEYER-GOHDE(Forthcoming): Solving DSGE Models with a Nonlinear Moving Average, Journal of
%               Economic Dynamics and Control.
%
%Included in the package pruning_abounds_1_0_1
%
%THIS VERSION: 1.0.9 April 29, 2013
%
%Copyright: Hong Lan and Alexander Meyer-Gohde
%
%You are free to use/modify/redistribute this program so long as original
%authorship credit is given and you in no way impinge on its free
%distribution
%This software is provided as is with no guarantees of any kind
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
global oo_
select_state=oo_.dr.nstatic+1:oo_.dr.nstatic+oo_.dr.npred;

if order == 1
   simulation_first=zeros(M_.endo_nbr,options_.periods);
   E=oo_.exo_simul';
   simulation_first(:,1)=oo_.dr.ghu*E(:,1); 
   for t=2:options_.periods
    simulation_first(:,t)=oo_.dr.ghx*simulation_first(select_state,t-1)+oo_.dr.ghu*E(:,t);
   end
   simulations.first=simulation_first(oo_.dr.inv_order_var,:);
   simulations.total=simulations.first+repmat(oo_.dr.ys,[1 options_.periods]);
end

if order==2
   simulation_first=zeros(M_.endo_nbr,options_.periods);
   simulation_second=zeros(M_.endo_nbr,options_.periods);
   if strcmp(type,'kim_et_al')
      E=oo_.exo_simul';
      simulation_first(:,1)=oo_.dr.ghu*E(:,1);
      exe=alt_kron(E(:,1),E(:,1));
      simulation_second(:,1)=(1/2)*(oo_.dr.ghuu*exe+oo_.dr.ghs2);
      for t=2:options_.periods
          simulation_first(:,t)=oo_.dr.ghx*simulation_first(select_state,t-1)+oo_.dr.ghu*E(:,t);
          exe=alt_kron(E(:,t),E(:,t));sxe=alt_kron(simulation_first(select_state,t-1),E(:,t));sxs=alt_kron(simulation_first(select_state,t-1),simulation_first(select_state,t-1));
          simulation_second(:,t)=oo_.dr.ghx*simulation_second(select_state,t-1)...
          +(1/2)*(oo_.dr.ghxx*sxs+2*oo_.dr.ghxu*sxe+oo_.dr.ghuu*exe+oo_.dr.ghs2);
      end
      simulations.first=simulation_first(oo_.dr.inv_order_var,:);
      simulations.second=simulation_second(oo_.dr.inv_order_var,:);
      simulations.total=simulations.second+simulations.first+repmat(oo_.dr.ys,[1 options_.periods]);
   end
   if strcmp(type,'den_haan_de_wind')
      E=oo_.exo_simul';
      simulation_first(:,1)=oo_.dr.ghu*E(:,1);
      exe=alt_kron(E(:,1),E(:,1));
      simulation_second(:,1)=(1/2)*oo_.dr.ghuu*exe;
      for t=2:options_.periods
          simulation_first(:,t)=oo_.dr.ghx*simulation_first(select_state,t-1)+oo_.dr.ghu*E(:,t);
          exe=alt_kron(E(:,t),E(:,t));sxe=alt_kron(simulation_first(select_state,t-1),E(:,t));sxs=alt_kron(simulation_first(select_state,t-1),simulation_first(select_state,t-1));
          simulation_second(:,t)=oo_.dr.ghx*simulation_second(select_state,t-1)...
          +(1/2)*(oo_.dr.ghxx*sxs+2*oo_.dr.ghxu*sxe+oo_.dr.ghuu*exe);
      end
      simulations.first=simulation_first(oo_.dr.inv_order_var,:);
      simulations.second=simulation_second(oo_.dr.inv_order_var,:);
      simulations.total=simulations.second+simulations.first+repmat(oo_.dr.ys+(1/2)*oo_.dr.ghs2(oo_.dr.inv_order_var),[1 options_.periods]);
   end
   if strcmp(type, 'lan_meyer-gohde')
      %compute nlma's y_sigma^2
        ghs2_state_nlma = (eye(oo_.dr.npred)-oo_.dr.ghx(oo_.dr.nstatic+1:oo_.dr.nstatic+oo_.dr.npred,:))\(oo_.dr.ghs2(oo_.dr.nstatic+1:oo_.dr.nstatic+oo_.dr.npred,:));
        ghs2_nlma = [ oo_.dr.ghx(1:oo_.dr.nstatic,:)*ghs2_state_nlma+oo_.dr.ghs2(1:oo_.dr.nstatic,:)
                      ghs2_state_nlma
                      oo_.dr.ghx(oo_.dr.nstatic+oo_.dr.npred+1:M_.endo_nbr,:)*ghs2_state_nlma+oo_.dr.ghs2(oo_.dr.nstatic+oo_.dr.npred+1:M_.endo_nbr,:)]; 
      %save results
        oo_.dr.ghs2_state_nlma = ghs2_state_nlma;
        oo_.dr.ghs2_nlma = ghs2_nlma; 
      E = oo_.exo_simul';
      simulation_first(:,1) = oo_.dr.ghu*E(:,1);
      exe = alt_kron(E(:,1),E(:,1));
      simulation_second(:,1) = (1/2)*oo_.dr.ghuu*exe;
      for t = 2:options_.periods
          simulation_first(:,t)=oo_.dr.ghx*simulation_first(select_state,t-1)+oo_.dr.ghu*E(:,t);
          exe = alt_kron(E(:,t),E(:,t));
          sxe = alt_kron(simulation_first(select_state,t-1),E(:,t));
          sxs = alt_kron(simulation_first(select_state,t-1),simulation_first(select_state,t-1));
          simulation_second(:,t) = oo_.dr.ghx*simulation_second(select_state,t-1)...
                                   +(1/2)*( oo_.dr.ghxx*sxs+2*oo_.dr.ghxu*sxe+oo_.dr.ghuu*exe );
      end
      simulations.first = simulation_first(oo_.dr.inv_order_var,:);
      simulations.second = simulation_second(oo_.dr.inv_order_var,:);
      simulations.total = simulations.second + simulations.first...
                          +repmat( oo_.dr.ys + 0.5*ghs2_nlma(oo_.dr.inv_order_var,:),[1 options_.periods] );
   end
end

if order==3 
   oo_=full_block_dr_new(oo_,M_,options_);%This seperates the risk correction of the first order coefficients from the first order coefficients themselves and seperates out the blocks of the third order policy function ghxxx etc.
   simulation_first=zeros(M_.endo_nbr,options_.periods);
   simulation_second=zeros(M_.endo_nbr,options_.periods);
   simulation_third=zeros(M_.endo_nbr,options_.periods);
   simulations.first_sigma_2=zeros(M_.endo_nbr,options_.periods);
   if strcmp(type,'andreasen')
      E=oo_.exo_simul';
      simulation_first(:,1)=oo_.dr.ghu*E(:,1);
      exe=alt_kron(E(:,1),E(:,1));
      simulation_second(:,1)=(1/2)*(oo_.dr.ghuu*exe+oo_.dr.ghs2);
      simulation_first_sigma_2(:,1)=(1/2)*(oo_.dr.ghs2u*E(:,1));
      simulation_third(:,1)=(1/6)*oo_.dr.ghuuu*alt_kron(E(:,1),exe);
      for t=2:options_.periods
          simulation_first(:,t)=oo_.dr.ghx*simulation_first(select_state,t-1)+oo_.dr.ghu*E(:,t);
          exe=alt_kron(E(:,t),E(:,t));sxe=alt_kron(simulation_first(select_state,t-1),E(:,t));sxs=alt_kron(simulation_first(select_state,t-1),simulation_first(select_state,t-1));
          simulation_second(:,t)=oo_.dr.ghx*simulation_second(select_state,t-1)...
              +(1/2)*(oo_.dr.ghxx*sxs+2*oo_.dr.ghxu*sxe+oo_.dr.ghuu*exe+oo_.dr.ghs2);
          simulation_first_sigma_2(:,t)=oo_.dr.ghx*simulation_first_sigma_2(select_state,t-1)+(1/2)*(oo_.dr.ghs2x*simulation_first(select_state,t-1)+oo_.dr.ghs2u*E(:,t));
          simulation_third(:,t)=oo_.dr.ghx*simulation_third(select_state,t-1)...
              +(1/6)*(oo_.dr.ghxxx*alt_kron(simulation_first(select_state,t-1),sxs)+oo_.dr.ghuuu*alt_kron(E(:,t),exe))...
              +(1/2)*(oo_.dr.ghxxu*alt_kron(sxs,E(:,t))+oo_.dr.ghxuu*alt_kron(simulation_first(select_state,t-1),exe))...
              +oo_.dr.ghxx*alt_kron(simulation_second(select_state,t-1),simulation_first(select_state,t-1))+oo_.dr.ghxu*alt_kron(simulation_second(select_state,t-1),E(:,t));
      end
      simulations.first=simulation_first(oo_.dr.inv_order_var,:);
      simulations.second=simulation_second(oo_.dr.inv_order_var,:);
      simulations.first_sigma_2=simulation_first_sigma_2(oo_.dr.inv_order_var,:);
      simulations.third=simulation_third(oo_.dr.inv_order_var,:);
      simulations.total=simulations.third+simulations.first_sigma_2+simulations.second+simulations.first+repmat(oo_.dr.ys,[1 options_.periods]);
   end
   if strcmp(type,'den_haan_de_wind')
      E=oo_.exo_simul';
      simulation_first(:,1)=(oo_.dr.ghu+(1/2)*oo_.dr.ghs2u)*E(:,1);
      exe=alt_kron(E(:,1),E(:,1));
      simulation_second(:,1)=(1/2)*oo_.dr.ghuu*exe;
      simulation_third(:,1)=(1/2)*oo_.dr.ghuu*exe+(1/6)*oo_.dr.ghuuu*alt_kron(E(:,1),exe);
      for t=2:options_.periods
          simulation_first(:,t)=(oo_.dr.ghx+(1/2)*oo_.dr.ghs2x)*simulation_first(select_state,t-1)+(oo_.dr.ghu+(1/2)*oo_.dr.ghs2u)*E(:,t);
          exe=alt_kron(E(:,t),E(:,t));sxe=alt_kron(simulation_first(select_state,t-1),E(:,t));sxs=alt_kron(simulation_first(select_state,t-1),simulation_first(select_state,t-1));
          simulation_second(:,t)=(oo_.dr.ghx+(1/2)*oo_.dr.ghs2x)*simulation_second(select_state,t-1)...
              +(1/2)*(oo_.dr.ghxx*sxs+2*oo_.dr.ghxu*sxe+oo_.dr.ghuu*exe);
          simulation_third(:,t)=(oo_.dr.ghx+(1/2)*oo_.dr.ghs2x)*simulation_third(select_state,t-1)+(oo_.dr.ghu+(1/2)*oo_.dr.ghs2u)*E(:,t)...
              +(1/2)*(oo_.dr.ghxx*alt_kron(simulation_second(select_state,t-1),simulation_second(select_state,t-1))+2*oo_.dr.ghxu*alt_kron(simulation_second(select_state,t-1),E(:,t))+oo_.dr.ghuu*exe)...
              +(1/6)*(oo_.dr.ghxxx*alt_kron(simulation_first(select_state,t-1),sxs)+oo_.dr.ghuuu*alt_kron(E(:,t),exe))...
              +(1/2)*(oo_.dr.ghxxu*alt_kron(sxs,E(:,t))+oo_.dr.ghxuu*alt_kron(simulation_first(select_state,t-1),exe));
      end
      simulations.first=simulation_first(oo_.dr.inv_order_var,:);
      simulations.second=simulation_second(oo_.dr.inv_order_var,:);
      simulations.third=simulation_third(oo_.dr.inv_order_var,:);
      simulations.total=simulations.third+repmat(oo_.dr.ys+(1/2)*oo_.dr.ghs2(oo_.dr.inv_order_var),[1 options_.periods]);
   end
   if strcmp(type,'fernandez-villaverde_et_al')
      E=oo_.exo_simul';
      simulation_first(:,1)=oo_.dr.ghu*E(:,1);
      exe=alt_kron(E(:,1),E(:,1));
      simulation_second(:,1)=(1/2)*(oo_.dr.ghuu*exe+oo_.dr.ghs2);
      simulation_first_sigma_2(:,1)=(1/2)*(oo_.dr.ghs2u*E(:,1));
      simulation_third(:,1)=(1/6)*oo_.dr.ghuuu*alt_kron(E(:,1),exe);
      for t=2:options_.periods
          simulation_first(:,t)=oo_.dr.ghx*simulation_first(select_state,t-1)+oo_.dr.ghu*E(:,t);
          exe=alt_kron(E(:,t),E(:,t));sxe=alt_kron(simulation_first(select_state,t-1),E(:,t));sxs=alt_kron(simulation_first(select_state,t-1),simulation_first(select_state,t-1));
          simulation_second(:,t)=oo_.dr.ghx*simulation_second(select_state,t-1)...
              +(1/2)*(oo_.dr.ghxx*sxs+2*oo_.dr.ghxu*sxe+oo_.dr.ghuu*exe+oo_.dr.ghs2);
          simulation_first_sigma_2(:,t)=oo_.dr.ghx*simulation_first_sigma_2(select_state,t-1)+(1/2)*(oo_.dr.ghs2x*simulation_first(select_state,t-1)+oo_.dr.ghs2u*E(:,t));
          simulation_third(:,t)=oo_.dr.ghx*simulation_third(select_state,t-1)...
              +(1/6)*(oo_.dr.ghxxx*alt_kron(simulation_first(select_state,t-1),sxs)+oo_.dr.ghuuu*alt_kron(E(:,t),exe))...
              +(1/2)*(oo_.dr.ghxxu*alt_kron(sxs,E(:,t))+oo_.dr.ghxuu*alt_kron(simulation_first(select_state,t-1),exe));
      end
      simulations.first=simulation_first(oo_.dr.inv_order_var,:);
      simulations.second=simulation_second(oo_.dr.inv_order_var,:);
      simulations.first_sigma_2=simulation_first_sigma_2(oo_.dr.inv_order_var,:);
      simulations.third=simulation_third(oo_.dr.inv_order_var,:);
      simulations.total=simulations.third+simulations.first_sigma_2+simulations.second+simulations.first+repmat(oo_.dr.ys,[1 options_.periods]);
   end
   if strcmp(type, 'lan_meyer-gohde')
      %compute nlma's y_sigma^2
        ghs2_state_nlma = (eye(oo_.dr.npred)-oo_.dr.ghx(oo_.dr.nstatic+1:oo_.dr.nstatic+oo_.dr.npred,:))\(oo_.dr.ghs2(oo_.dr.nstatic+1:oo_.dr.nstatic+oo_.dr.npred,:));
        ghs2_nlma = [ oo_.dr.ghx(1:oo_.dr.nstatic,:)*ghs2_state_nlma+oo_.dr.ghs2(1:oo_.dr.nstatic,:)
                      ghs2_state_nlma
                      oo_.dr.ghx(oo_.dr.nstatic+oo_.dr.npred+1:M_.endo_nbr,:)*ghs2_state_nlma+oo_.dr.ghs2(oo_.dr.nstatic+oo_.dr.npred+1:M_.endo_nbr,:)]; 
      %compute nlma's y_sigma^2e and y_sigma^2y^state
        %y_sigma^2e           
           ghs2u_nlma = oo_.dr.ghs2u + oo_.dr.ghxu*alt_kron(ghs2_state_nlma,eye(M_.exo_nbr));
        %y_sigma^2y^state
           ghs2x_nlma = oo_.dr.ghs2x + oo_.dr.ghxx*alt_kron(ghs2_state_nlma,eye(oo_.dr.npred));
      %save results
        oo_.dr.ghs2_state_nlma = ghs2_state_nlma;
        oo_.dr.ghs2_nlma = ghs2_nlma;
        oo_.dr.ghs2u_nlma = ghs2u_nlma;
        oo_.dr.ghs2x_nlma = ghs2x_nlma;
      E = oo_.exo_simul';
      simulation_first(:,1) = oo_.dr.ghu*E(:,1);
      exe = alt_kron(E(:,1),E(:,1));
      simulation_second(:,1) = (1/2)*oo_.dr.ghuu*exe;      
      simulation_third = (1/6)*oo_.dr.ghuuu*alt_kron(E(:,1),exe);      
      simulation_first_sigma_2(:,1) = (1/2)*(ghs2u_nlma*E(:,1));
      for t = 2:options_.periods
          simulation_first(:,t)=oo_.dr.ghx*simulation_first(select_state,t-1)+oo_.dr.ghu*E(:,t);
          exe = alt_kron(E(:,t),E(:,t));
          sxe = alt_kron(simulation_first(select_state,t-1),E(:,t));
          sxs = alt_kron(simulation_first(select_state,t-1),simulation_first(select_state,t-1));
          simulation_second(:,t) = oo_.dr.ghx*simulation_second(select_state,t-1)...
                                   +(1/2)*( oo_.dr.ghxx*sxs+2*oo_.dr.ghxu*sxe+oo_.dr.ghuu*exe );
          simulation_first_sigma_2(:,t) = oo_.dr.ghx*simulation_first_sigma_2(select_state,t-1)...
                                          +(1/2)*(ghs2u_nlma*E(:,t)+ghs2x_nlma*simulation_first(select_state,t-1));
          simulation_third(:,t) = oo_.dr.ghx*simulation_third(select_state,t-1)...
                                  +(1/6)*(oo_.dr.ghxxx*alt_kron(simulation_first(select_state,t-1),sxs)+oo_.dr.ghuuu*alt_kron(E(:,t),exe))...
                                  +(1/2)*(oo_.dr.ghxxu*alt_kron(sxs,E(:,t))+oo_.dr.ghxuu*alt_kron(simulation_first(select_state,t-1),exe))...
                                  +oo_.dr.ghxx*alt_kron(simulation_second(select_state,t-1),simulation_first(select_state,t-1))...
                                  +oo_.dr.ghxu*alt_kron(simulation_second(select_state,t-1),E(:,t));
      end
      simulations.first = simulation_first(oo_.dr.inv_order_var,:);
      simulations.second = simulation_second(oo_.dr.inv_order_var,:);      
      simulations.first_sigma_2 = simulation_first_sigma_2(oo_.dr.inv_order_var,:);
      simulations.third = simulation_third(oo_.dr.inv_order_var,:);
      simulations.total = simulations.third +simulations.first_sigma_2 + simulations.second + simulations.first...
                          +repmat( oo_.dr.ys + 0.5*ghs2_nlma(oo_.dr.inv_order_var,:),[1 options_.periods] );
   end
end








